@@ -1,3 +1,36 @@
+/*
+ * Copyright (c) 2013 Genome Research Ltd.
+ * Author(s): James Bonfield, Rob Davies
+ * 
+ * Redistribution and use in source and binary forms, with or without 
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *    1. Redistributions of source code must retain the above copyright notice,
+ *       this list of conditions and the following disclaimer.
+ * 
+ *    2. Redistributions in binary form must reproduce the above
+ *       copyright notice, this list of conditions and the following
+ *       disclaimer in the documentation and/or other materials provided
+ *       with the distribution.
+ * 
+ *    3. Neither the names Genome Research Ltd and Wellcome Trust Sanger
+ *    Institute nor the names of its contributors may be used to endorse
+ *    or promote products derived from this software without specific
+ *    prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY GENOME RESEARCH LTD AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GENOME RESEARCH
+ * LTD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
 /*
  * Author: James Bonfield, Wellcome Trust Sanger Institute. 2010-3
  */
@@ -427,9 +460,7 @@ bam_file_t *bam_open(const char *fn, const char *mode) {
 
     /* Load first block so we can check */
     bam_more_input(b);
-    if (b->comp_sz < 2)
-	return NULL;
-    if (b->comp_p[0] == 31 && b->comp_p[1] == 139)
+    if (b->comp_sz >= 2 && b->comp_p[0] == 31 && b->comp_p[1] == 139)
 	b->gzip = 1;
     else
 	b->gzip = 0;
@@ -445,13 +476,13 @@ bam_file_t *bam_open(const char *fn, const char *mode) {
     if (-1 == bam_uncompress_input(b))
 	return NULL;
     /* Auto-correct open file type if we detect a BAM */
-    if (b->uncomp_sz >= 3 && strncmp("BAM", (char *)b->uncomp_p, 3) != 0) {
-	b->mode &= ~O_BINARY;
-	mode = "r";
-    } else if (b->uncomp_sz >= 3 && strncmp("BAM", (char *)b->uncomp_p, 3) == 0) {
+    if (b->uncomp_sz >= 4 && strncmp("BAM\001", (char *)b->uncomp_p, 4) == 0) {
 	b->mode |= O_BINARY;
 	b->binary = 1;
 	mode = "rb";
+    } else {
+	b->mode &= ~O_BINARY;
+	mode = "r";
     }
 
     /* Load header */
