@@ -1,3 +1,36 @@
+/*
+ * Copyright (c) 2013 Genome Research Ltd.
+ * Author(s): James Bonfield
+ * 
+ * Redistribution and use in source and binary forms, with or without 
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *    1. Redistributions of source code must retain the above copyright notice,
+ *       this list of conditions and the following disclaimer.
+ * 
+ *    2. Redistributions in binary form must reproduce the above
+ *       copyright notice, this list of conditions and the following
+ *       disclaimer in the documentation and/or other materials provided
+ *       with the distribution.
+ * 
+ *    3. Neither the names Genome Research Ltd and Wellcome Trust Sanger
+ *    Institute nor the names of its contributors may be used to endorse
+ *    or promote products derived from this software without specific
+ *    prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY GENOME RESEARCH LTD AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GENOME RESEARCH
+ * LTD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
 /*
  * Author: James Bonfield, Wellcome Trust Sanger Institute. 2013
  */
@@ -21,6 +54,9 @@
 #include "io_lib/os.h"
 #include "io_lib/md5.h"
 
+#ifdef SAMTOOLS
+#    define bam_copy(dst, src) bam_copy1(*(dst), (src))
+#else
 void bam_copy(bam_seq_t **bt, bam_seq_t *bf) {
     size_t a;
 
@@ -35,8 +71,11 @@ void bam_copy(bam_seq_t **bt, bam_seq_t *bf) {
 
     (*bt)->alloc = a;
 }
+#endif
 
 #define Z_CRAM_STRAT Z_FILTERED
+//#define Z_CRAM_STRAT Z_RLE
+//#define Z_CRAM_STRAT Z_HUFFMAN_ONLY
 //#define Z_CRAM_STRAT Z_DEFAULT_STRATEGY
 
 static int process_one_read(cram_fd *fd, cram_container *c,
@@ -711,7 +750,7 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 
 	//printf("BF=0x%x\n", cr->flags);
 	//	    bf = cram_flag_swap[cr->flags];
-	i32 = fd->cram_flag_swap[cr->flags & 0x7ff];
+	i32 = fd->cram_flag_swap[cr->flags & 0xfff];
 	r |= h->BF_codec->encode(s, h->BF_codec, core, (char *)&i32, 1);
 
 	i32 = cr->cram_flags;
@@ -969,28 +1008,49 @@ static int cram_encode_slice(cram_fd *fd, cram_container *c,
 #endif
 
     /* Compress the other blocks */
-    if (cram_compress_block(fd, s->block[1], NULL,
+    if (cram_compress_block(fd, s->block[1], NULL, //IN (seq)
 			    fd->level, Z_CRAM_STRAT,
 			    -1, -1))
 	return -1;
-    if (cram_compress_block(fd, s->block[2], fd->m[1],
-			    fd->level, Z_CRAM_STRAT, 
-			    LEVEL2, STRAT2))
-	return -1;
-    if (cram_compress_block(fd, s->block[3], NULL,
+
+    if (fd->level == 0) {
+	/* Do nothing */
+    } else if (fd->level == 1) {
+	if (cram_compress_block(fd, s->block[2], fd->m[1], //qual
+				1, Z_RLE, -1, -1))
+	    return -1;
+	if (cram_compress_block(fd, s->block[5], fd->m[4], //Tags
+				1, Z_RLE, -1, -1))
+	    return -1;
+    } else if (fd->level < 3) {
+	if (cram_compress_block(fd, s->block[2], fd->m[1], //qual
+				1, Z_RLE,
+				1, Z_HUFFMAN_ONLY))
+	    return -1;
+	if (cram_compress_block(fd, s->block[5], fd->m[4], //Tags
+				1, Z_RLE,
+				1, Z_HUFFMAN_ONLY))
+	    return -1;
+    } else {
+	if (cram_compress_block(fd, s->block[2], fd->m[1], //qual
+				fd->level, Z_CRAM_STRAT, 
+				LEVEL2, STRAT2))
+	    return -1;
+	if (cram_compress_block(fd, s->block[5], fd->m[4], //Tags
+				fd->level, Z_CRAM_STRAT,
+				LEVEL2, STRAT2))
+	    return -1;
+    }
+    if (cram_compress_block(fd, s->block[3], NULL, //Name
 			    fd->level, Z_CRAM_STRAT,
 			    -1, -1))
 	return -1;
-    if (cram_compress_block(fd, s->block[4], NULL,
+    if (cram_compress_block(fd, s->block[4], NULL, //TS, NP
 			    fd->level, Z_CRAM_STRAT,
 			    -1, -1))
 	return -1;
-    if (cram_compress_block(fd, s->block[5], fd->m[4],
-			    fd->level, Z_CRAM_STRAT,
-	                    LEVEL2, STRAT2))
-	return -1;
     if (fd->version != CRAM_1_VERS) {
-	if (cram_compress_block(fd, s->block[6], NULL,
+	if (cram_compress_block(fd, s->block[6], NULL, //SC (seq)
 				fd->level, Z_CRAM_STRAT,
 				-1, -1))
 	    return -1;
@@ -1027,14 +1087,19 @@ int cram_encode_container(cram_fd *fd, cram_container *c) {
     cram_block_compression_hdr *h = c->comp_hdr;
     cram_block *c_hdr;
     int multi_ref = 0;
-    int r1, r2, sn;
+    int r1, r2, sn, nref;
     spare_bams *spares;
 
     /* Cache references up-front if we have unsorted access patterns */
+    pthread_mutex_lock(&fd->ref_lock);
+    nref = fd->refs->nref;
+    pthread_mutex_unlock(&fd->ref_lock);
+
     if (c->refs_used) {
-	for (i = 0; i < fd->refs->nref; i++) {
-	    if (c->refs_used[i])
+	for (i = 0; i < nref; i++) {
+	    if (c->refs_used[i]) {	
 		cram_get_ref(fd, i, 1, 0);
+	    }
 	}
     }
 
@@ -1801,7 +1866,7 @@ static char *cram_encode_aux(cram_fd *fd, bam_seq_t *b, cram_container *c,
 			     cram_slice *s, cram_record *cr) {
     char *aux, *orig, *tmp, *rg = NULL;
 #ifdef SAMTOOLS
-    int aux_size = b->l_aux;
+    int aux_size = bam_get_l_aux(b);
 #else
     int aux_size = bam_blk_size(b) -
 	((char *)bam_aux(b) - (char *)&bam_ref(b));
@@ -2103,7 +2168,7 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     cr->flags       = bam_flag(b);
     if (bam_cigar_len(b) == 0)
 	cr->flags |= BAM_FUNMAP;
-    cram_stats_add(c->BF_stats, fd->cram_flag_swap[cr->flags & 0x7ff]);
+    cram_stats_add(c->BF_stats, fd->cram_flag_swap[cr->flags & 0xfff]);
 
     if (!fd->no_ref)
 	cr->cram_flags = CRAM_FLAG_PRESERVE_QUAL_SCORES;
@@ -2130,7 +2195,22 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     cr->name_len    = bam_name_len(b);
     cram_stats_add(c->RN_stats, cr->name_len);
 
+#if 1
     BLOCK_APPEND(s->name_blk, bam_name(b), bam_name_len(b));
+#else
+    // Experiment with using delta encoding to last name
+    {
+	int l = bam_name_len(b);
+	char *n1 = bam_name(b), *n0 = c->last_name;
+	for (i = 0; i < l; i++) {
+	    if (n1[i] != n0[i])
+		break;
+	}
+	BLOCK_APPEND_CHAR(s->name_blk, i);
+	BLOCK_APPEND(s->name_blk, bam_name(b)+i, bam_name_len(b)-i);
+	c->last_name = n1;
+    }
+#endif
 
 
     /*
@@ -2147,7 +2227,11 @@ static int process_one_read(cram_fd *fd, cram_container *c,
     *seq = 0;
     for (i = 0; i < cr->len; i++) {
 	// FIXME: do 2 char at a time for efficiency
+#ifdef SAMTOOLS
+	cp[i] = seq_nt16_str[bam_seqi(bam_seq(b), i)];
+#else
 	cp[i] = bam_nt16_rev_table[bam_seqi(bam_seq(b), i)];
+#endif
     }
     BLOCK_SIZE(s->seqs_blk) += cr->len;
 
@@ -2303,14 +2387,15 @@ static int process_one_read(cram_fd *fd, cram_container *c,
 	if (cr->len == 0) {
 	    cram_stats_add(c->RL_stats, cr->len = fake_qual);
 	    BLOCK_GROW(s->qual_blk, cr->len);
-	    qual = cp = (char *)BLOCK_END(s->qual_blk);
+	    cp = (char *)BLOCK_END(s->qual_blk);
 	    memset(cp, 255, cr->len);
 	} else {
 	    BLOCK_GROW(s->qual_blk, cr->len);
-	    qual = cp = (char *)BLOCK_END(s->qual_blk);
-	    for (i = 0; i < cr->len; i++) {
-		cp[i] = bam_qual(b)[i];
-	    }
+	    cp = (char *)BLOCK_END(s->qual_blk);
+	    char *from = &bam_qual(b)[0];
+	    char *to = &cp[0];
+	    memcpy(to, from, cr->len);
+	    //for (i = 0; i < cr->len; i++) cp[i] = from[i];
 	}
 	BLOCK_SIZE(s->qual_blk) += cr->len;
     } else {
@@ -2485,7 +2570,9 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	    c->pos_sorted = 0; // required atm for multi_seq slices
 
 	    if (!c->refs_used) {
+		pthread_mutex_lock(&fd->ref_lock);
 		c->refs_used = calloc(fd->refs->nref, sizeof(int));
+		pthread_mutex_unlock(&fd->ref_lock);
 		if (!c->refs_used)
 		    return -1;
 	    }
@@ -2499,12 +2586,16 @@ int cram_put_bam_seq(cram_fd *fd, bam_seq_t *b) {
 	    !fd->unsorted) {
 	    
 	    if (!c->refs_used) {
+		pthread_mutex_lock(&fd->ref_lock);
 		c->refs_used = calloc(fd->refs->nref, sizeof(int));
+		pthread_mutex_unlock(&fd->ref_lock);
 		if (!c->refs_used)
 		    return -1;
 	    } else if (c->refs_used && c->refs_used[bam_ref(b)]) {
 		fprintf(stderr, "Unsorted mode enabled\n");
+		pthread_mutex_lock(&fd->ref_lock);
 		fd->unsorted = 1;
+		pthread_mutex_unlock(&fd->ref_lock);
 		fd->multi_seq = 1;
 	    }
 	}
