@@ -1,3 +1,36 @@
+/*
+ * Copyright (c) 2013 Genome Research Ltd.
+ * Author(s): James Bonfield
+ * 
+ * Redistribution and use in source and binary forms, with or without 
+ * modification, are permitted provided that the following conditions are met:
+ * 
+ *    1. Redistributions of source code must retain the above copyright notice,
+ *       this list of conditions and the following disclaimer.
+ * 
+ *    2. Redistributions in binary form must reproduce the above
+ *       copyright notice, this list of conditions and the following
+ *       disclaimer in the documentation and/or other materials provided
+ *       with the distribution.
+ * 
+ *    3. Neither the names Genome Research Ltd and Wellcome Trust Sanger
+ *    Institute nor the names of its contributors may be used to endorse
+ *    or promote products derived from this software without specific
+ *    prior written permission.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY GENOME RESEARCH LTD AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GENOME RESEARCH
+ * LTD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
+
 /*
  * Author: James Bonfield, Wellcome Trust Sanger Institute. 2013
  *
@@ -50,6 +83,7 @@
 #include <sys/syscall.h>
 #define gettid() (int)syscall(SYS_gettid)
 
+
 /* ----------------------------------------------------------------------
  * ITF8 encoding and decoding.
  *
@@ -512,11 +546,14 @@ static char *zlib_mem_inflate(char *cdata, size_t csize, size_t *size) {
     err = inflateInit2(&s, 15 + 32);
     if (err != Z_OK) {
 	fprintf(stderr, "zlib inflateInit error: %s\n", s.msg);
+	free(data);
 	return NULL;
     }
 
     /* Decode to 'data' array */
     for (;s.avail_in;) {
+	unsigned char *data_tmp;
+
 	s.next_out = &data[s.total_out];
 	err = inflate(&s, Z_NO_FLUSH);
 	if (err == Z_STREAM_END)
@@ -528,9 +565,11 @@ static char *zlib_mem_inflate(char *cdata, size_t csize, size_t *size) {
 	}
 
 	/* More to come, so realloc */
-	data = realloc(data, data_alloc += s.avail_in + 10);
-	if (!data)
+	data = realloc((data_tmp = data), data_alloc += s.avail_in + 10);
+	if (!data) {
+	    free(data_tmp);
 	    return NULL;
+	}
 	s.avail_out += s.avail_in+10;
     }
     inflateEnd(&s);
@@ -734,7 +773,8 @@ int cram_uncompress_block(cram_block *b) {
 	uncomp = zlib_mem_inflate((char *)b->data, b->comp_size, &uncomp_size);
 	if (!uncomp)
 	    return -1;
-	assert((int)uncomp_size == b->uncomp_size);
+	if ((int)uncomp_size != b->uncomp_size)
+	    return -1;
 	free(b->data);
 	b->data = (unsigned char *)uncomp;
 	b->method = RAW;
@@ -760,9 +800,12 @@ int cram_uncompress_block(cram_block *b) {
     case BZIP2:
 	fprintf(stderr, "Bzip2 compression is not compiled into this "
 		"version.\nPlease rebuild and try again.\n");
-	abort();
+	return -1;
 	break;
 #endif
+
+    default:
+	return -1;
     }
 
     return 0;
@@ -831,6 +874,8 @@ int cram_compress_block(cram_fd *fd, cram_block *b, cram_metrics *metrics,
 	// metrics ignored for bzip2
 	return cram_compress_block_bzip2(fd, b, metrics, level);
 #endif
+    
+    //return cram_compress_block_arith1(fd, b, metrics);
 
     pthread_mutex_lock(&fd->metrics_lock);
     if (strat2 >= 0)
@@ -1144,21 +1189,33 @@ static refs_t *refs_load_fai(refs_t *r_orig, char *fn, int is_err) {
 	e->seq = NULL;
 
 	hd.p = e;
-	if (!(hi = HashTableAdd(r->h_meta, e->name, strlen(e->name), hd, &n)))
+	if (!(hi = HashTableAdd(r->h_meta, e->name, strlen(e->name), hd, &n))){
+	    free(e);
 	    return NULL;
+	}
 
 	if (!n) {
-	    /* Replace old one if needed */
+	    /* Replace old one if needed. */
 	    ref_entry *r = (ref_entry *)hi->data.p;
-	    if (r)
-		free(r);
-	    hi->data.p = e;
+
+	    if (r && (r->count != 0 || r->length != 0)) {
+		/* Keep old one */
+		free(e);
+	    } else {
+		if (r)
+		    free(r);
+		hi->data.p = e;
+	    }	    
 	}
     }
 
+    RP("refs_load_fai %s END (success)\n", fn);
+
     return r;
 
  err:
+    RP("refs_load_fai %s END (fail)\n", fn);
+
     if (fp)
 	fclose(fp);
 
@@ -1177,8 +1234,11 @@ static refs_t *refs_load_fai(refs_t *r_orig, char *fn, int is_err) {
  */
 int refs2id(refs_t *r, SAM_hdr *h) {
     int i;
+
     if (r->ref_id)
 	free(r->ref_id);
+    if (r->last)
+	r->last = NULL;
 
     r->ref_id = calloc(h->nref, sizeof(*r->ref_id));
     if (!r->ref_id)
@@ -1235,6 +1295,9 @@ static int refs_from_header(refs_t *r, cram_fd *fd, SAM_hdr *h) {
 	if (!(r->ref_id[i] = calloc(1, sizeof(ref_entry))))
 	    return -1;
 
+	if (!h->ref[i].name)
+	    return -1;
+
 	r->ref_id[i]->name = string_dup(r->pool, h->ref[i].name);
 	r->ref_id[i]->length = 0; // marker for not yet loaded
 
@@ -1268,7 +1331,6 @@ void expand_cache_path(char *path, char *dir, char *fn) {
     while ((cp = strchr(dir, '%'))) {
 	strncpy(path, dir, cp-dir);
 	path += cp-dir;
-	dir += cp-dir;
 
 	if (*++cp == 's') {
 	    strcpy(path, fn);
@@ -1473,7 +1535,7 @@ static int cram_populate_ref(cram_fd *fd, int id, ref_entry *r) {
 }
 
 static void cram_ref_incr_locked(refs_t *r, int id) {
-    RP("%d INC REF %d, %d\n", gettid(), id, (int)(id>=0?r->ref_id[id]->count+1:-1));
+    RP("%d INC REF %d, %d %p\n", gettid(), id, (int)(id>=0?r->ref_id[id]->count+1:-999), id>=0?r->ref_id[id]->seq:(char *)1);
 
     if (id < 0 || !r->ref_id[id]->seq)
 	return;
@@ -1487,13 +1549,11 @@ void cram_ref_incr(refs_t *r, int id) {
     pthread_mutex_unlock(&r->lock);
 }
 
-void cram_ref_decr(refs_t *r, int id) {
-    pthread_mutex_lock(&r->lock);
-
-    RP("%d DEC REF %d, %d\n", gettid(), id, (int)(id>=0?r->ref_id[id]->count-1:-1));
+static void cram_ref_decr_locked(refs_t *r, int id) {
+    RP("%d DEC REF %d, %d %p\n", gettid(), id, (int)(id>=0?r->ref_id[id]->count-1:-999), id>=0?r->ref_id[id]->seq:(char *)1);
 
     if (id < 0 || !r->ref_id[id]->seq) {
-	pthread_mutex_unlock(&r->lock);
+	assert(r->ref_id[id]->count >= 0);
 	return;
     }
 
@@ -1503,9 +1563,14 @@ void cram_ref_decr(refs_t *r, int id) {
 	if (r->ref_id[id]->seq) {
 	    free(r->ref_id[id]->seq);
 	    r->ref_id[id]->seq = NULL;
+	    r->ref_id[id]->length = 0;
 	}
     }
+}
 
+void cram_ref_decr(refs_t *r, int id) {
+    pthread_mutex_lock(&r->lock);
+    cram_ref_decr_locked(r, id);
     pthread_mutex_unlock(&r->lock);
 }
 
@@ -1522,6 +1587,9 @@ static char *load_ref_portion(FILE *fp, ref_entry *e, int start, int end) {
     off_t offset, len;
     char *seq;
 
+    if (end < start)
+	end = start;
+
     /*
      * Compute locations in file. This is trivial for the MD5 files, but
      * is still necessary for the fasta variants.
@@ -1541,12 +1609,13 @@ static char *load_ref_portion(FILE *fp, ref_entry *e, int start, int end) {
 	return NULL;
     }
 
-    if (!(seq = malloc(len))) {
+    if (len == 0 || !(seq = malloc(len))) {
 	return NULL;
     }
 
     if (len != fread(seq, 1, len, fp)) {
 	perror("fread() on reference file");
+	free(seq);
 	return NULL;
     }
 
@@ -1564,6 +1633,7 @@ static char *load_ref_portion(FILE *fp, ref_entry *e, int start, int end) {
 
 	if (cp_to - seq != end-start+1) {
 	    fprintf(stderr, "Malformed reference file?\n");
+	    free(seq);
 	    return NULL;
 	}
     }
@@ -1583,16 +1653,24 @@ ref_entry *cram_ref_load(refs_t *r, int id) {
     int start = 1, end = e->length;
     char *seq;
 
-    if (e->seq)
+    if (e->seq) {
 	return e;
+    }
 
     assert(e->count == 0);
 
     if (r->last) {
-	RP("%d cram_ref_load DECR %d\n", gettid(), r->last - r->ref_id[0]);
+#ifdef REF_DEBUG
+	int idx = 0;
+	for (idx = 0; idx < r->nref; idx++)
+	    if (r->last == r->ref_id[idx])
+		break;
+	RP("%d cram_ref_load DECR %d => %d\n", gettid(), idx, r->last->count-1);
+#endif
+
 	assert(r->last->count > 0);
 	if (--r->last->count <= 0) {
-	    RP("%d FREE REF %d (%p)\n", gettid(), id, r->ref_id[id]->seq);
+	    RP("%d FREE REF %d (%p)\n", gettid(), id, r->last->seq);
 	    if (r->last->seq) {
 		free(r->last->seq);
 		r->last->seq = NULL;
@@ -1627,7 +1705,7 @@ ref_entry *cram_ref_load(refs_t *r, int id) {
      * Also keep track of last used ref so incr/decr loops on the same
      * sequence don't cause load/free loops.
      */
-    RP("%d cram_ref_load INCR %d\n", gettid(), id);
+    RP("%d cram_ref_load INCR %d => %d\n", gettid(), id, e->count+1);
     r->last = e;
     e->count++; 
 
@@ -1709,16 +1787,21 @@ char *cram_get_ref(cram_fd *fd, int id, int start, int end) {
      * A ref entry computed from @SQ lines (M5 or UR field) will have
      * r->length == 0 unless it's been loaded once and verified that we have
      * an on-disk filename for it.
+     *
+     * 19 Sep 2013: Moved the lock here as the cram_populate_ref code calls
+     * open_path_mfile and libcurl, which isn't multi-thread safe unless I
+     * rewrite my code to have one curl handle per thread.
      */
+    pthread_mutex_lock(&fd->refs->lock);
     if (r->length == 0) {
 	if (cram_populate_ref(fd, id, r) == -1) {
 	    fprintf(stderr, "Failed to populate reference for id %d\n", id);
+	    pthread_mutex_unlock(&fd->refs->lock);
 	    pthread_mutex_unlock(&fd->ref_lock);
 	    return NULL;
 	}
 	r = fd->refs->ref_id[id];
     }
-    pthread_mutex_lock(&fd->refs->lock);
 
 
     /*
@@ -1766,7 +1849,7 @@ char *cram_get_ref(cram_fd *fd, int id, int start, int end) {
 		    cram_ref_incr_locked(fd->refs, id);
 	    }	    
 
-	    fd->ref       = r->seq;
+	    fd->ref = NULL; /* We never access it directly */
 	    fd->ref_start = 1;
 	    fd->ref_end   = r->length;
 	    fd->ref_id    = id;
@@ -1900,47 +1983,58 @@ cram_container *cram_new_container(int nrec, int nslice) {
     c->bams = NULL;
 
     if (!(c->slices = (cram_slice **)calloc(nslice, sizeof(cram_slice *))))
-	return NULL;
+	goto err;
     c->slice = NULL;
 
     if (!(c->comp_hdr = cram_new_compression_header()))
-	return NULL;
+	goto err;
     c->comp_hdr_block = NULL;
 
-    if (!(c->BF_stats = cram_stats_create())) return NULL;
-    if (!(c->CF_stats = cram_stats_create())) return NULL;
-    if (!(c->RN_stats = cram_stats_create())) return NULL;
-    if (!(c->AP_stats = cram_stats_create())) return NULL;
-    if (!(c->RG_stats = cram_stats_create())) return NULL;
-    if (!(c->MQ_stats = cram_stats_create())) return NULL;
-    if (!(c->NS_stats = cram_stats_create())) return NULL;
-    if (!(c->NP_stats = cram_stats_create())) return NULL;
-    if (!(c->TS_stats = cram_stats_create())) return NULL;
-    if (!(c->MF_stats = cram_stats_create())) return NULL;
-    if (!(c->NF_stats = cram_stats_create())) return NULL;
-    if (!(c->RL_stats = cram_stats_create())) return NULL;
-    if (!(c->FN_stats = cram_stats_create())) return NULL;
-    if (!(c->FC_stats = cram_stats_create())) return NULL;
-    if (!(c->FP_stats = cram_stats_create())) return NULL;
-    if (!(c->DL_stats = cram_stats_create())) return NULL;
-    if (!(c->BA_stats = cram_stats_create())) return NULL;
-    if (!(c->QS_stats = cram_stats_create())) return NULL;
-    if (!(c->BS_stats = cram_stats_create())) return NULL;
-    if (!(c->TC_stats = cram_stats_create())) return NULL;
-    if (!(c->TN_stats = cram_stats_create())) return NULL;
-    if (!(c->TL_stats = cram_stats_create())) return NULL;
-    if (!(c->RI_stats = cram_stats_create())) return NULL;
-    if (!(c->RS_stats = cram_stats_create())) return NULL;
-    if (!(c->PD_stats = cram_stats_create())) return NULL;
-    if (!(c->HC_stats = cram_stats_create())) return NULL;
+    if (!(c->BF_stats = cram_stats_create())) goto err;
+    if (!(c->CF_stats = cram_stats_create())) goto err;
+    if (!(c->RN_stats = cram_stats_create())) goto err;
+    if (!(c->AP_stats = cram_stats_create())) goto err;
+    if (!(c->RG_stats = cram_stats_create())) goto err;
+    if (!(c->MQ_stats = cram_stats_create())) goto err;
+    if (!(c->NS_stats = cram_stats_create())) goto err;
+    if (!(c->NP_stats = cram_stats_create())) goto err;
+    if (!(c->TS_stats = cram_stats_create())) goto err;
+    if (!(c->MF_stats = cram_stats_create())) goto err;
+    if (!(c->NF_stats = cram_stats_create())) goto err;
+    if (!(c->RL_stats = cram_stats_create())) goto err;
+    if (!(c->FN_stats = cram_stats_create())) goto err;
+    if (!(c->FC_stats = cram_stats_create())) goto err;
+    if (!(c->FP_stats = cram_stats_create())) goto err;
+    if (!(c->DL_stats = cram_stats_create())) goto err;
+    if (!(c->BA_stats = cram_stats_create())) goto err;
+    if (!(c->QS_stats = cram_stats_create())) goto err;
+    if (!(c->BS_stats = cram_stats_create())) goto err;
+    if (!(c->TC_stats = cram_stats_create())) goto err;
+    if (!(c->TN_stats = cram_stats_create())) goto err;
+    if (!(c->TL_stats = cram_stats_create())) goto err;
+    if (!(c->RI_stats = cram_stats_create())) goto err;
+    if (!(c->RS_stats = cram_stats_create())) goto err;
+    if (!(c->PD_stats = cram_stats_create())) goto err;
+    if (!(c->HC_stats = cram_stats_create())) goto err;
     
     //c->aux_B_stats = cram_stats_create();
 
     if (!(c->tags_used = HashTableCreate(16, HASH_DYNAMIC_SIZE)))
-	return NULL;
+	goto err;
     c->refs_used = 0;
 
+    c->last_name = "";
+
     return c;
+
+ err:
+    if (c) {
+	if (c->slices)
+	    free(c->slices);
+	free(c);
+    }
+    return NULL;
+
 }
 
 void cram_free_container(cram_container *c) {
@@ -2272,11 +2366,16 @@ cram_block_compression_hdr *cram_new_compression_header(void) {
     if (!hdr)
 	return NULL;
 
-    if (!(hdr->TD_blk = cram_new_block(CORE, 0)))
+    if (!(hdr->TD_blk = cram_new_block(CORE, 0))) {
+	free(hdr);
 	return NULL;
+    }
 
-    if (!(hdr->TD = HashTableCreate(16, HASH_DYNAMIC_SIZE)))
+    if (!(hdr->TD = HashTableCreate(16, HASH_DYNAMIC_SIZE))) {
+	cram_free_block(hdr->TD_blk);
+	free(hdr);
 	return NULL;
+    }
 
     return hdr;
 }
@@ -2447,7 +2546,7 @@ cram_slice *cram_new_slice(enum cram_content_type type, int nrecs) {
 	return NULL;
 
     if (!(s->hdr = (cram_block_slice_hdr *)calloc(1, sizeof(*s->hdr))))
-	return NULL;
+	goto err;
     s->hdr->content_type = type;
 
     s->hdr_block = NULL;
@@ -2455,19 +2554,19 @@ cram_slice *cram_new_slice(enum cram_content_type type, int nrecs) {
     s->block_by_id = NULL;
     s->last_apos = 0;
     s->id = 0;
-    if (!(s->crecs = malloc(nrecs * sizeof(cram_record))))        return NULL;
+    if (!(s->crecs = malloc(nrecs * sizeof(cram_record))))        goto err;
     s->cigar = NULL;
     s->cigar_alloc = 0;
     s->ncigar = 0;
 
-    if (!(s->seqs_blk = cram_new_block(EXTERNAL, 0)))             return NULL;
-    if (!(s->qual_blk = cram_new_block(EXTERNAL, CRAM_EXT_QUAL))) return NULL;
-    if (!(s->name_blk = cram_new_block(EXTERNAL, CRAM_EXT_NAME))) return NULL;
-    if (!(s->aux_blk  = cram_new_block(EXTERNAL, CRAM_EXT_TAG)))  return NULL;
-    if (!(s->base_blk = cram_new_block(EXTERNAL, CRAM_EXT_IN)))   return NULL;
-    if (!(s->soft_blk = cram_new_block(EXTERNAL, CRAM_EXT_SC)))   return NULL;
+    if (!(s->seqs_blk = cram_new_block(EXTERNAL, 0)))             goto err;
+    if (!(s->qual_blk = cram_new_block(EXTERNAL, CRAM_EXT_QUAL))) goto err;
+    if (!(s->name_blk = cram_new_block(EXTERNAL, CRAM_EXT_NAME))) goto err;
+    if (!(s->aux_blk  = cram_new_block(EXTERNAL, CRAM_EXT_TAG)))  goto err;
+    if (!(s->base_blk = cram_new_block(EXTERNAL, CRAM_EXT_IN)))   goto err;
+    if (!(s->soft_blk = cram_new_block(EXTERNAL, CRAM_EXT_SC)))   goto err;
 #ifdef TN_external
-    if (!(s->tn_blk   = cram_new_block(EXTERNAL, CRAM_EXT_TN)))   return NULL;
+    if (!(s->tn_blk   = cram_new_block(EXTERNAL, CRAM_EXT_TN)))   goto err;
 #endif
 
     s->features = NULL;
@@ -2479,13 +2578,19 @@ cram_slice *cram_new_slice(enum cram_content_type type, int nrecs) {
 #endif
 
     // Volatile keys as we do realloc in dstring
-    if (!(s->pair = HashTableCreate(10000, HASH_DYNAMIC_SIZE)))   return NULL;
+    if (!(s->pair = HashTableCreate(10000, HASH_DYNAMIC_SIZE)))   goto err;
     
 #ifdef BA_external
     s->BA_len = 0;
 #endif
 
     return s;
+
+ err:
+    if (s)
+	cram_free_slice(s);
+
+    return NULL;
 }
 
 /*
@@ -2502,7 +2607,7 @@ cram_slice *cram_new_slice(enum cram_content_type type, int nrecs) {
 cram_slice *cram_read_slice(cram_fd *fd) {
     cram_block *b = cram_read_block(fd);
     cram_slice *s = calloc(1, sizeof(*s));
-    int i, n, max_id;
+    int i, n, max_id, min_id;
 
     if (!b || !s)
 	goto err;
@@ -2525,16 +2630,19 @@ cram_slice *cram_read_slice(cram_fd *fd) {
     if (!s->block)
 	goto err;
 
-    for (max_id = i = 0; i < n; i++) {
+    for (max_id = i = 0, min_id = INT_MAX; i < n; i++) {
 	if (!(s->block[i] = cram_read_block(fd)))
 	    goto err;
 
-	if (s->block[i]->content_type == EXTERNAL &&
-	    max_id < s->block[i]->content_id)
-	    max_id = s->block[i]->content_id;
+	if (s->block[i]->content_type == EXTERNAL) {
+	    if (max_id < s->block[i]->content_id)
+		max_id = s->block[i]->content_id;
+	    if (min_id > s->block[i]->content_id)
+		min_id = s->block[i]->content_id;
+	}
     }
-    if (max_id < 1024) {
-	if (!(s->block_by_id = calloc(max_id+1, sizeof(s->block[0]))))
+    if (min_id >= 0 && max_id < 1024) {
+	if (!(s->block_by_id = calloc(1024, sizeof(s->block[0]))))
 	    goto err;
 
 	for (i = 0; i < n; i++) {
@@ -2571,8 +2679,10 @@ cram_slice *cram_read_slice(cram_fd *fd) {
  err:
     if (b)
 	cram_free_block(b);
-    if (s)
+    if (s) {
+	s->hdr_block = NULL;
 	cram_free_slice(s);
+    }
     return NULL;
 }
 
@@ -2733,7 +2843,11 @@ SAM_hdr *cram_read_SAM_hdr(cram_fd *fd) {
     }
 
     /* Parse */
+#ifdef SAMTOOLS
+    hdr = sam_hdr_parse_(header, header_len);
+#else
     hdr = sam_hdr_parse(header, header_len);
+#endif
     free(header);
 
     return hdr;
@@ -2788,6 +2902,7 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 	int i;
 	for (i = 0; i < hdr->nref; i++) {
 	    SAM_hdr_type *ty;
+	    char *ref;
 
 	    if (!(ty = sam_hdr_find(hdr, "SQ", "SN", hdr->ref[i].name)))
 		return -1;
@@ -2799,8 +2914,8 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 
 		rlen = fd->refs->ref_id[i]->length;
 		MD5_Init(&md5);
-		cram_get_ref(fd, i, 1, rlen);
-		MD5_Update(&md5, fd->ref, rlen);
+		ref = cram_get_ref(fd, i, 1, rlen);
+		MD5_Update(&md5, ref, rlen);
 		MD5_Final(buf, &md5);
 
 		for (j = 0; j < 16; j++) {
@@ -2870,8 +2985,11 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 
 	c->num_blocks = 2;
 	c->num_landmarks = 2;
-	if (!(c->landmark = malloc(2*sizeof(*c->landmark))))
+	if (!(c->landmark = malloc(2*sizeof(*c->landmark)))) {
+	    cram_free_block(b);
+	    cram_free_container(c);
 	    return -1;
+	}
 	c->landmark[0] = 0;
 	c->landmark[1] = c->length;
 
@@ -2881,8 +2999,11 @@ int cram_write_SAM_hdr(cram_fd *fd, SAM_hdr *hdr) {
 #ifdef PADDED_BLOCK
 	padded_length = MAX(c->length*2, 10000) - c->length;
 	c->length += padded_length;
-	if (NULL == (pads = calloc(1, padded_length)))
+	if (NULL == (pads = calloc(1, padded_length))) {
+	    cram_free_block(b);
+	    cram_free_container(c);
 	    return -1;
+	}
 	BLOCK_APPEND(b, pads, padded_length);
 	BLOCK_UPLEN(b);
 	free(pads);
@@ -2981,7 +3102,7 @@ static void cram_init_tables(cram_fd *fd) {
 	    fd->bam_flag_swap[i]  = f;
 	}
     
-	for (i = 0; i < 0x800; i++) {
+	for (i = 0; i < 0x1000; i++) {
 	    int g = 0;
 
 	    if (i & BAM_FPAIRED)	   g |= CRAM_FPAIRED;
@@ -2998,9 +3119,9 @@ static void cram_init_tables(cram_fd *fd) {
 	}
     } else {
 	/* NOP */
-	for (i = 0; i < 0x800; i++)
+	for (i = 0; i < 0x1000; i++)
 	    fd->bam_flag_swap[i] = i;
-	for (i = 0; i < 0x800; i++)
+	for (i = 0; i < 0x1000; i++)
 	    fd->cram_flag_swap[i] = i;
     }
 
@@ -3126,6 +3247,7 @@ cram_fd *cram_open(const char *filename, const char *mode) {
     fd->shared_ref = 0;
 
     fd->index       = NULL;
+    fd->own_pool    = 0;
     fd->pool        = NULL;
     fd->rqueue      = NULL;
     fd->job_pending = NULL;
@@ -3251,6 +3373,9 @@ int cram_close(cram_fd *fd) {
     if (fd->index)
 	cram_index_free(fd);
 
+    if (fd->own_pool && fd->pool)
+	t_pool_destroy(fd->pool, 0);
+
     free(fd);
     return 0;
 }
@@ -3363,6 +3488,22 @@ int cram_set_voption(cram_fd *fd, enum cram_option opt, va_list args) {
 	fd->multi_seq = va_arg(args, int);
 	break;
 
+    case CRAM_OPT_NTHREADS: {
+	int nthreads =  va_arg(args, int);
+        if (nthreads > 1) {
+            if (!(fd->pool = t_pool_init(nthreads*2, nthreads)))
+                return -1;
+
+	    fd->rqueue = t_results_queue_init();
+	    pthread_mutex_init(&fd->metrics_lock, NULL);
+	    pthread_mutex_init(&fd->ref_lock, NULL);
+	    pthread_mutex_init(&fd->bam_list_lock, NULL);
+	    fd->shared_ref = 1;
+	    fd->own_pool = 1;
+        }
+	break;
+    }
+
     case CRAM_OPT_THREAD_POOL:
 	fd->pool = va_arg(args, t_pool *);
 	if (fd->pool) {
@@ -3372,6 +3513,7 @@ int cram_set_voption(cram_fd *fd, enum cram_option opt, va_list args) {
 	    pthread_mutex_init(&fd->bam_list_lock, NULL);
 	}
 	fd->shared_ref = 1; // Needed to avoid clobbering ref between threads
+	fd->own_pool = 0;
 
 	//fd->qsize = 1;
 	//fd->decoded = calloc(fd->qsize, sizeof(cram_container *));
